/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.registry.app;

import org.apache.abdera.Abdera;
import org.apache.abdera.model.*;
import org.apache.abdera.protocol.Response;
import org.apache.abdera.protocol.client.AbderaClient;
import org.apache.abdera.protocol.client.ClientResponse;
import org.apache.abdera.protocol.client.RequestOptions;
import org.apache.axiom.om.util.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.Comment;
import org.wso2.registry.*;
import org.wso2.registry.exceptions.ResourceNotFoundException;
import org.wso2.registry.i18n.Messages;

import javax.xml.namespace.QName;
import java.io.*;
import java.net.URL;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class RemoteRegistry implements Registry, RegistryConstants {

    private String baseURI;
    private Log log = LogFactory.getLog(RemoteRegistry.class);

    /**
     * To create a remote registry need to provide a URL of a remote registry and the URL should be
     * something like http://localhost:8080/wso2registry/atom/r1
     *
     * @param registryURL : URL to the registry or to the resource
     * @throws RegistryException : If something went wrong
     */
    public RemoteRegistry(URL registryURL) throws RegistryException {
        baseURI = registryURL.toString();
    }

    public Resource get(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + path);
        if (clientResponse.getType() == Response.ResponseType.CLIENT_ERROR) {
            throw new ResourceNotFoundException(path);
        }
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String state = feed.getSimpleExtension(new QName("state"));
        if (state != null && state.equals("Deleted")) {
            throw new ResourceNotFoundException(path);
        }
        return createResourceFromFeed(feed, false);
    }

    /**
     * This method will generate a resource object representing the Feed object
     * and the logic will be simply the reverse of the create feed of the Atom
     * registry
     *
     * @param feed    : Feed object which represent a resource object
     * @param isQuery : To tell whether we are going to execute a query of trying to
     *                just get a resource
     * @return : Created resource
     */
    private Resource createResourceFromFeed(Feed feed, boolean isQuery) {
        String isDirectory = feed.getSimpleExtension(new QName("directory"));
        Resource resource = new Resource();
        List attributes = feed.getAttributes();
        for (Object attribute : attributes) {
            QName attName = (QName) attribute;
            String key = attName.getLocalPart();
            resource.setProperty(key, feed.getAttributeValue(key));
        }
        if (feed.getAuthor() != null) {
            resource.setAuthorUserName(feed.getAuthor().getName());
        }
        resource.setLastModified(feed.getUpdated());
        String cretedDate = feed.getSimpleExtension(new QName("createdTime"));
        if (cretedDate != null) {
            resource.setCreatedTime(new Date(Long.parseLong(cretedDate)));
        }
        resource.setPath(feed.getTitle());
        resource.setDescription(feed.getSubtitle());
        String state = feed.getSimpleExtension(new QName("state"));
        if (state != null && "Deleted".equals(state)) {
            resource.setState(DELETED_STATE);
        }

        if (isDirectory != null && "true".equals(isDirectory)) {
            resource.setDirectory(true);
            List entries = feed.getEntries();
            if (entries != null) {
                String[] childNodes = new String[entries.size()];
                for (int i = 0; i < entries.size(); i++) {
                    Entry entry = (Entry) entries.get(i);
                    if (isQuery) {
                        childNodes[i] = entry.getLink("queryResult").getHref().toString();
                    } else {
                        childNodes[i] = entry.getLink("path").getHref().toString();
                    }

                }
                resource.setContent(childNodes);
            }
        } else {
            List entries = feed.getEntries();
            // if the resource is not a directory then we have only one entry
            Entry entry = (Entry) entries.get(0);
            String content1 = entry.getContent();
            if (content1 != null && !Content.Type.TEXT.equals(entry.getContentType())) {
                byte[] content = Base64.decode(content1);
                resource.setContent(content);
            } else {
                resource.setContent(content1);
            }
        }
        return resource;
    }

    public boolean resourceExists(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + path +
                URL_SEPARATOR +
                PARAMETER_RESOURCE_EXIST);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String boolValue = feed.getSubtitle();
        return "True".equals(boolValue);
    }

    public void put(String path, Resource resource) throws RegistryException {
        // Figure out where we're POSTing to...

        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        if (SQL_QUERY_MEDIA_TYPE.equals(resource.getMediaType())) {
            entry.setAttributeValue("mediaType", SQL_QUERY_MEDIA_TYPE);
        }
        entry.setSummary(resource.getDescription());
        entry.addAuthor(resource.getAuthorUserName());
        Properties properties = resource.getProperties();
        if (properties != null) {
            Iterator keys = properties.keySet().iterator();
            while (keys.hasNext()) {
                String key = (String) keys.next();
                entry.setAttributeValue(key, (String) properties.get(key));
            }
        }
        if (resource.isDirectory()) {
            entry.addSimpleExtension(new QName("directory"), "true");
        } else {
            entry.addSimpleExtension(new QName("directory"), "false");
            Object content = resource.getContent();
            if (content != null && content instanceof byte[]) {
                ByteArrayInputStream in = new ByteArrayInputStream((byte[]) content);
                entry.setContent(in);
            } else if (content instanceof InputStream) {
                entry.setContent((InputStream) content);
            } else {
                entry.setContent((String) content);
            }
        }
        ClientResponse resp = abderaClient.post(baseURI + path, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.add", path));
        } else {
            log.error(Messages.getMessage("add.resource.fail", path));
            throw new RegistryException(Messages.getMessage("add.resource.fail", path));
        }
    }

    public void delete(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse resp = abderaClient.delete(baseURI + path);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.deleted", path));
        } else {
            log.error(Messages.getMessage("resource.delete.fail", path));
            throw new RegistryException(Messages.getMessage("resource.delete.fail", path));
        }
    }

    public String[] getVersions(String path) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + path +
                URL_SEPARATOR +
                PARAMETER_VERSION);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        if (entries != null) {
            String[] versions = new String[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry) entries.get(i);
                versions[i] = entry.getLink("versionLink").getHref().toString();
            }
            return versions;
        }
        return new String[0];
    }

    public void restoreVersion(String versionPath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        ClientResponse resp = abderaClient.put(baseURI + versionPath +
                URL_SEPARATOR +
                PARAMETER_RESTORE, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.restoreded", versionPath));
        } else {
            log.error(Messages.getMessage("resource.restore.fail", versionPath));
            throw new RegistryException(Messages.getMessage("resource.restore.fail", versionPath));
        }
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setContent(tag);
        ClientResponse resp = abderaClient.put(baseURI + resourcePath +
                URL_SEPARATOR +
                PARAMETER_TAGS, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.tagged", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.tag.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.tag.fails", resourcePath));
        }
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI +
                URL_SEPARATOR +
                PARAMETER_TAG_PATHS +
                ":" + tag);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        TaggedResourcePath taggedResourcePaths[] = null;
        if (entries != null) {
            taggedResourcePaths = new TaggedResourcePath[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry) entries.get(i);
                TaggedResourcePath tagpath = new TaggedResourcePath();
                tagpath.setResourcePath(entry.getTitle());
                tagpath.setTagCount(Long.parseLong(entry.getSimpleExtension(new QName("taggings"))));
                taggedResourcePaths[i] = tagpath;
            }
        }
        return taggedResourcePaths;
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI +
                resourcePath +
                URL_SEPARATOR +
                PARAMETER_TAGS);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        Tag tags[] = null;
        if (entries != null) {
            tags = new Tag[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry) entries.get(i);
                Tag tag = new Tag();
                tag.setTagCount(Long.parseLong(entry.getSimpleExtension(new QName("taggings"))));
                tag.setTagName(entry.getTitle());
                tags[i] = tag;
            }
        }
        return tags;
    }

    public void removeTag(String path, String tag) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setContent(tag);
        ClientResponse resp = abderaClient.put(baseURI + path +
                URL_SEPARATOR +
                PARAMETER_TAGS_REMOVED, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("tag.removed", path));
        } else {
            log.error(Messages.getMessage("tag.removed.fails", path));
            throw new RegistryException(Messages.getMessage("tag.removed.fails", path));
        }
    }

    public void addComment(String resourcePath, Comment comment) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setUpdated(comment.getCommentedTime());
        entry.addAuthor(comment.getCommentedUser());
        entry.setContent(comment.getCommentText());
        ClientResponse resp = abderaClient.put(baseURI + resourcePath +
                URL_SEPARATOR +
                PARAMETER_COMMENTS, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.commented", resourcePath));
        } else {
            log.error(Messages.getMessage("resouce.comment.fails", resourcePath));
            throw new RegistryException(Messages.getMessage("resouce.comment.fails", resourcePath));
        }
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + resourcePath +
                URL_SEPARATOR +
                PARAMETER_COMMENTS);
        Document introspection =
                clientResponse.getDocument();
        Element element = introspection.getRoot();
        Feed feed = (Feed) element;
        List entries = feed.getEntries();
        Comment comments[] = null;
        if (entries != null) {
            comments = new Comment[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry) entries.get(i);
                Comment comment = new Comment();
                comment.setCommentedTime(entry.getUpdated());
                comment.setCommentText(entry.getContent());
                comment.setCommentedUser(entry.getAuthor().getName());
                comments[i] = comment;
            }
        }
        return comments;
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        Entry entry = abdera.getFactory().newEntry();
        entry.setContent("" + rating);
        entry.setSummary(PARAMETER_RATINGS);
        ClientResponse resp = abderaClient.put(baseURI + resourcePath +
                URL_SEPARATOR +
                PARAMETER_RATINGS, entry);
        if (resp.getType() == Response.ResponseType.SUCCESS) {
            log.info(Messages.getMessage("resource.rated", resourcePath));
        } else {
            log.error(Messages.getMessage("resource.rate.fail", resourcePath));
            throw new RegistryException(Messages.getMessage("resource.rate.fail", resourcePath));
        }
    }

    public float getAverageRating(String resourcePath) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse = abderaClient.get(baseURI + resourcePath +
                URL_SEPARATOR + PARAMETER_RATINGS);
        Document introspection = clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String floatvalue = feed.getSimpleExtension(QNAME_AVGRATING);
        return Float.parseFloat(floatvalue);
    }

    public int getRating(String path, String userName) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        ClientResponse clientResponse =
                abderaClient.get(baseURI + path + URL_SEPARATOR +
                                 PARAMETER_RATINGS + ":" + userName);
        Document introspection =
                clientResponse.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        String intvalue = feed.getSubtitle();
        return Integer.parseInt(intvalue);
    }


    public Resource executeQuery(String path, Object[] parameters) throws RegistryException {
        Abdera abdera = new Abdera();
        AbderaClient abderaClient = new AbderaClient(abdera);
        RequestOptions requestOptions = new RequestOptions();
        requestOptions.addHeader("parameters", encodeParametesAsString(parameters));
        ClientResponse resp = abderaClient.get(baseURI + path +
                URL_SEPARATOR +
                PARAMETER_QUERY, requestOptions);
        Document introspection =
                resp.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        return createResourceFromFeed(feed, true);
    }

    public static String encodeParametesAsString(Object[] parameters) {
        String value = "";
        for (int i = 0; i < parameters.length; i++) {
            Object parameter = parameters[i];
            if (i < parameters.length - 1) {
                value = value + parameter + "#";
            } else {
                value = value + parameter;
            }
        }
        return value;
    }

    public static String[] decodeParametesAsString(String value) {
        return value.split("\\#");
    }

    public LogEntry[] getLogs(String resourcePath,
                              int action,
                              String userName,
                              Date from,
                              Date to,
                              boolean recentFirst) throws RegistryException {

        Abdera abdera = new Abdera();
        if (resourcePath == null || "".equals(resourcePath)) {
            resourcePath = "/";
        }
        AbderaClient abderaClient = new AbderaClient(abdera);
        RequestOptions requestOptions = new RequestOptions();
        requestOptions.addDateHeader("ToDate", to);
        requestOptions.addDateHeader("FromDate", from);
        requestOptions.addHeader("Action", "" + action);
        requestOptions.addHeader("Author", userName);
        ClientResponse resp = abderaClient.get(baseURI + resourcePath +
                URL_SEPARATOR + PARAMETER_LOGS, requestOptions);
        Document introspection =
                resp.getDocument();
        Feed feed = (Feed) introspection.getRoot();
        List entries = feed.getEntries();
        LogEntry logs[] = null;
        if (entries != null) {
            logs = new LogEntry[entries.size()];
            for (int i = 0; i < entries.size(); i++) {
                Entry entry = (Entry) entries.get(i);
                LogEntry logEntry = new LogEntry();
                logEntry.setDate(entry.getEdited());
                logEntry.setActionData(entry.getContent());
                logEntry.setUserName(entry.getAuthor().getName());
                logEntry.setAction(Integer.parseInt(entry.getSimpleExtension(new QName("action"))));
                logEntry.setResourcePath(entry.getSimpleExtension(new QName("path")));
                logs[i] = logEntry;
            }
        }
        return logs;
    }

    /**
     * This method can be used to export a local file system into a running instance of a registry.
     * Need to create a file object representing the local file and the need to tell where to add
     * the resource in the registry.
     *
     * @param file : File representing local file system
     * @param path : Where to put the file
     * @throws RegistryException : if something went wrong
     */
    public void importToRegistry(File file, String path) throws RegistryException {
        try {
            if (file == null || path == null) {
                throw new RegistryException(Messages.getMessage("file.valaue.null"));
            }
            processImport(file, path);
        } catch (Exception e) {
            throw new RegistryException(e.getMessage());
        }
    }

    /**
     * This method can be used to export registry instance or node in a registry to a local file
     * system. When we call this method a file structure will be created to match the structure to
     * map the structure in the registry
     *
     * @param toFile : File in the local file system
     * @param path   : To which node export
     * @throws RegistryException : If something went wrong
     */
    public void exportFromRegistry(File toFile, String path) throws RegistryException {
        try {
            processExport(path, toFile);
        } catch (Exception e) {
            e.printStackTrace();
            throw new RegistryException(e.getMessage());
        }
    }

    private void processImport(File file, String path) throws Exception {
        String filePart = file.getName();
        String resourcePath = path + "/" + filePart;
        if (file.isDirectory()) {
            File files[] = file.listFiles();
            if (files.length > 0) {
                for (File childFile : files) {
                    processImport(childFile, resourcePath);
                }
            } else {
                Resource resource = new Resource();
                resource.setPath(resourcePath);
                resource.setDirectory(true);
                put(resourcePath, resource);
            }
        } else {
            Resource resource = new Resource();
            resource.setContent(new FileInputStream(file));
            resource.setPath(resourcePath);
            resource.setDirectory(false);
            put(resourcePath, resource);
        }
    }

    private void processExport(String resourcePath,
                               File toFile) throws RegistryException {
        Resource resource = get(resourcePath);
        if (resource != null) {
            if (resource.isDirectory()) {
                File file = new File(toFile, resource.getPath());
                file.mkdirs();
                String childNodes[] = (String[]) resource.getContent();
                for (String childNode : childNodes) {
                    processExport(childNode, toFile);
                }
            } else {
                File file = new File(toFile, resource.getPath());
                try {
                    file.createNewFile();
                    FileOutputStream out = new FileOutputStream(file);
                    out.write((byte[]) resource.getContent());
                    out.flush();
                    out.close();
                } catch (IOException e) {
                    throw new RegistryException(Messages.getMessage(
                            "error.creating.file", resource.getPath()));
                }

            }
        }
    }
}
